from typing import Any, List, Mapping, Optional, Union

from chalk.features import Feature
from chalk.sql.finalized_query import FinalizedChalkQuery, Finalizer
from chalk.sql.protocols import BaseSQLSourceProtocol, ChalkQueryProtocol, IncrementalSettings
from chalk.utils.duration import Duration, parse_chalk_duration
from chalk.utils.missing_dependency import missing_dependency_exception


class ChalkQuery(ChalkQueryProtocol):
    def __init__(
        self,
        features: Mapping[str, Feature],
        targets: List[Any],
        source: BaseSQLSourceProtocol,
    ):
        try:
            from sqlalchemy.orm import Query
        except ImportError:
            raise missing_dependency_exception("chalkpy[sql]")
        self._query = Query(entities=targets)
        self._features = features
        self._source = source

    def first(self, params: Optional[Mapping[str, Any]] = None):
        from sqlalchemy.sql import Select

        self._query = self._query.limit(1)
        assert isinstance(self._query.selectable, Select), "A text query's selectable should always be a Select"
        return FinalizedChalkQuery(
            query=self._query.selectable,
            finalizer=Finalizer.FIRST,
            source=self._source,
            incremental_settings=None,
            params=params or {},
            fields=self._features,
        )

    def one_or_none(self, params: Optional[Mapping[str, Any]] = None):
        from sqlalchemy.sql import Select

        self._query = self._query.limit(1)
        assert isinstance(self._query.selectable, Select), "A text query's selectable should always be a Select"
        return FinalizedChalkQuery(
            query=self._query.selectable,
            finalizer=Finalizer.ONE_OR_NONE,
            source=self._source,
            incremental_settings=None,
            params=params or {},
            fields=self._features,
        )

    def one(self, params: Optional[Mapping[str, Any]] = None):
        from sqlalchemy.sql import Select

        self._query = self._query.limit(1)
        assert isinstance(self._query.selectable, Select), "A text query's selectable should always be a Select"
        return FinalizedChalkQuery(
            query=self._query.selectable,
            finalizer=Finalizer.ONE,
            source=self._source,
            incremental_settings=None,
            params=params or {},
            fields=self._features,
        )

    def all(self, params: Optional[Mapping[str, Any]] = None):
        from sqlalchemy.sql import Select

        assert isinstance(self._query.selectable, Select), "A text query's selectable should always be a Select"
        return FinalizedChalkQuery(
            query=self._query.selectable,
            finalizer=Finalizer.ALL,
            source=self._source,
            incremental_settings=None,
            params=params or {},
            fields=self._features,
        )

    def incremental(
        self,
        lookback_period: Duration = "0s",
        mode: str = "row",
        incremental_column: Optional[Union[str, Feature]] = None,
        params: Optional[Mapping[str, Any]] = None,
    ):
        from sqlalchemy.sql import Select

        assert isinstance(self._query.selectable, Select)
        return FinalizedChalkQuery(
            query=self._query.selectable,
            finalizer=Finalizer.ALL,
            source=self._source,
            incremental_settings=IncrementalSettings(
                lookback_period=(
                    parse_chalk_duration(lookback_period) if isinstance(lookback_period, str) else lookback_period
                ),
                incremental_column=None if incremental_column is None else str(incremental_column),
                mode=mode,
            ),
            params=params or {},
            fields=self._features,
        )

    def filter_by(self, **kwargs: Any):
        self._query = self._query.filter_by(**kwargs)
        return self

    def filter(self, *criterion: Any):
        self._query = self._query.filter(*criterion)
        return self

    def limit(self, *limits: Any):
        self._query = self._query.limit(*limits)
        return self

    def order_by(self, *clauses: Any):
        self._query = self._query.order_by(*clauses)
        return self

    def group_by(self, *clauses: Any):
        self._query = self._query.group_by(*clauses)
        return self

    def having(self, criterion: Any):
        self._query = self._query.having(*criterion)
        return self

    def union(self, *q: Any):
        self._query = self._query.union(*q)
        return self

    def union_all(self, *q: Any):
        self._query = self._query.union_all(*q)
        return self

    def intersect(self, *q: Any):
        self._query = self._query.intersect(*q)
        return self

    def intersect_all(self, *q: Any):
        self._query = self._query.intersect_all(*q)
        return self

    def join(self, target: Any, *props: Any, **kwargs: Any):
        self._query = self._query.join(target, *props, **kwargs)
        return self

    def outerjoin(self, target: Any, *props: Any, **kwargs: Any):
        self._query = self._query.outerjoin(target, *props, **kwargs)
        return self

    def select_from(self, *from_obj: Any):
        self._query = self._query.select_from(*from_obj)
        return self



# AUTO-GENERATED FILE. Do not edit.
from enum import Enum


class MetricKind(str, Enum):
    FEATURE_REQUEST_COUNT = "FEATURE_REQUEST_COUNT"
    FEATURE_LATENCY = "FEATURE_LATENCY"
    FEATURE_STALENESS = "FEATURE_STALENESS"
    FEATURE_VALUE = "FEATURE_VALUE"
    FEATURE_WRITE = "FEATURE_WRITE"
    FEATURE_NULL_RATIO = "FEATURE_NULL_RATIO"

    RESOLVER_REQUEST_COUNT = "RESOLVER_REQUEST_COUNT"
    RESOLVER_LATENCY = "RESOLVER_LATENCY"
    RESOLVER_SUCCESS_RATIO = "RESOLVER_SUCCESS_RATIO"

    QUERY_COUNT = "QUERY_COUNT"
    QUERY_LATENCY = "QUERY_LATENCY"
    QUERY_SUCCESS_RATIO = "QUERY_SUCCESS_RATIO"

    BILLING_INFERENCE = "BILLING_INFERENCE"
    BILLING_CRON = "BILLING_CRON"
    BILLING_MIGRATION = "BILLING_MIGRATION"

    CRON_COUNT = "CRON_COUNT"
    CRON_LATENCY = "CRON_LATENCY"

    STREAM_MESSAGES_PROCESSED = "STREAM_MESSAGES_PROCESSED"
    STREAM_MESSAGE_LATENCY = "STREAM_MESSAGE_LATENCY"

    STREAM_WINDOWS_PROCESSED = "STREAM_WINDOWS_PROCESSED"
    STREAM_WINDOW_LATENCY = "STREAM_WINDOW_LATENCY"


class FilterKind(str, Enum):
    FEATURE_STATUS = "FEATURE_STATUS"
    FEATURE_NAME = "FEATURE_NAME"
    FEATURE_TAG = "FEATURE_TAG"

    RESOLVER_STATUS = "RESOLVER_STATUS"
    RESOLVER_NAME = "RESOLVER_NAME"
    RESOLVER_TAG = "RESOLVER_TAG"

    CRON_STATUS = "CRON_STATUS"
    MIGRATION_STATUS = "MIGRATION_STATUS"

    ONLINE_OFFLINE = "ONLINE_OFFLINE"
    CACHE_HIT = "CACHE_HIT"
    OPERATION_ID = "OPERATION_ID"

    QUERY_NAME = "QUERY_NAME"
    QUERY_STATUS = "QUERY_STATUS"

    IS_NULL = "IS_NULL"


class ComparatorKind(str, Enum):
    EQ = "EQ"
    NEQ = "NEQ"
    ONE_OF = "ONE_OF"


class WindowFunctionKind(str, Enum):
    COUNT = "COUNT"
    MEAN = "MEAN"
    SUM = "SUM"
    MIN = "MIN"
    MAX = "MAX"

    PERCENTILE_99 = "PERCENTILE_99"
    PERCENTILE_95 = "PERCENTILE_95"
    PERCENTILE_75 = "PERCENTILE_75"
    PERCENTILE_50 = "PERCENTILE_50"
    PERCENTILE_25 = "PERCENTILE_25"
    PERCENTILE_5 = "PERCENTILE_5"

    ALL_PERCENTILES = "ALL_PERCENTILES"

    @classmethod
    def has_member_key(cls, key):
        return key in cls.__members__


class GroupByKind(str, Enum):
    FEATURE_STATUS = "FEATURE_STATUS"
    FEATURE_NAME = "FEATURE_NAME"
    IS_NULL = "IS_NULL"

    RESOLVER_STATUS = "RESOLVER_STATUS"
    RESOLVER_NAME = "RESOLVER_NAME"

    QUERY_STATUS = "QUERY_STATUS"
    QUERY_NAME = "QUERY_NAME"

    ONLINE_OFFLINE = "ONLINE_OFFLINE"
    CACHE_HIT = "CACHE_HIT"


class ChartLinkKind(str, Enum):
    resolver = "resolver"
    feature = "feature"
    query = "query"
    manual = "manual"


class MetricFormulaKind(str, Enum):
    SUM = "SUM"
    TOTAL_RATIO = "TOTAL_RATIO"
    RATIO = "RATIO"
    PRODUCT = "PRODUCT"
    ABS = "ABS"
    KS_STAT = "KS_STAT"
    KS_TEST = "KS_TEST"
    KS_THRESHOLD = "KS_THRESHOLD"
    TIME_OFFSET = "TIME_OFFSET"


class AlertSeverityKind(str, Enum):
    CRITICAL = "CRITICAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"


class ThresholdPosition(str, Enum):
    ABOVE = "ABOVE"
    BELOW = "BELOW"



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterator, List, Optional

from chalk.clogging import chalk_logger
from chalk.features import FeatureConverter
from chalk.integrations.named import load_integration_variable
from chalk.sql._internal.sql_source import BaseSQLSource, validate_dtypes_for_efficient_execution
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import pyarrow as pa
    from snowflake.connector.result_batch import ResultBatch
    from sqlalchemy.engine.url import URL


try:
    import sqlalchemy as sa
except ImportError:
    sa = None

if sa is None:
    _supported_sqlalchemy_types_for_pa_querying = ()
else:
    _supported_sqlalchemy_types_for_pa_querying = (
        sa.BigInteger,
        sa.Boolean,
        sa.BINARY,
        sa.BLOB,
        sa.LargeBinary,
        sa.Float,
        sa.Integer,
        sa.Time,
        sa.String,
        sa.Text,
        sa.VARBINARY,
        sa.DateTime,
        sa.Date,
        sa.SmallInteger,
        sa.BIGINT,
        sa.BOOLEAN,
        sa.CHAR,
        sa.DATETIME,
        sa.FLOAT,
        sa.INTEGER,
        sa.SMALLINT,
        sa.TEXT,
        sa.TIMESTAMP,
        sa.VARCHAR,
    )


class SnowflakeSourceImpl(BaseSQLSource):
    def __init__(
        self,
        *,
        name: Optional[str] = None,
        account_identifier: Optional[str] = None,
        warehouse: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        db: Optional[str] = None,
        schema: Optional[str] = None,
        role: Optional[str] = None,
    ):
        try:
            import snowflake  # noqa
            import snowflake.sqlalchemy  # noqa
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[snowflake]")
        del snowflake  # unused

        self.account_identifier = account_identifier or load_integration_variable(
            integration_name=name, name="SNOWFLAKE_ACCOUNT_ID"
        )
        self.warehouse = warehouse or load_integration_variable(integration_name=name, name="SNOWFLAKE_WAREHOUSE")
        self.user = user or load_integration_variable(integration_name=name, name="SNOWFLAKE_USER")
        self.password = password or load_integration_variable(integration_name=name, name="SNOWFLAKE_PASSWORD")
        self.db = db or load_integration_variable(integration_name=name, name="SNOWFLAKE_DATABASE")
        self.schema = schema or load_integration_variable(integration_name=name, name="SNOWFLAKE_SCHEMA")
        self.role = role or load_integration_variable(integration_name=name, name="SNOWFLAKE_ROLE")
        BaseSQLSource.__init__(self, name=name)

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        query = {
            k: v
            for k, v in (
                {
                    "database": self.db,
                    "schema": self.schema,
                    "warehouse": self.warehouse,
                    "role": self.role,
                }
            ).items()
            if v is not None
        }
        return URL.create(
            drivername="snowflake",
            username=self.user,
            password=self.password,
            host=self.account_identifier,
            query=query,
        )

    def execute_query_efficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
        connection: Optional[Any] = None,
    ):
        # these imports are safe because the only way we end up here is if we have a valid SnowflakeSource constructed,
        # which already gates this import
        import pyarrow as pa
        import snowflake.connector
        from sqlalchemy.sql import Select

        if isinstance(finalized_query.query, Select):
            validate_dtypes_for_efficient_execution(finalized_query.query, _supported_sqlalchemy_types_for_pa_querying)

        with (
            snowflake.connector.connect(
                user=self.user,
                account=self.account_identifier,
                password=self.password,
                warehouse=self.warehouse,
                schema=self.schema,
                database=self.db,
            )
            if connection is None
            else contextlib.nullcontext(connection)
        ) as con:
            if con is None:
                chalk_logger.warning("Failed to connect to Snowflake")
            assert con is not None
            chalk_logger.info("Established connection with Snowflake")
            sql, positional_params, named_params = self.compile_query(finalized_query)
            assert len(positional_params) == 0, "using named param style"
            with con.cursor() as cursor:
                chalk_logger.info("Acquired cursor for Snowflake query. Executing.")
                res = cursor.execute(sql, named_params)
                chalk_logger.info("Executed Snowflake query. Fetching results.")
                chalk_logger.info(f"Compiled query: {sql}")
                assert res is not None

                chalk_logger.info("Fetching arrow tables from Snowflake.")
                arrows = cursor.fetch_arrow_all()

                if arrows is None:
                    chalk_logger.info("Query returned no results.")
                    return cursor.get_result_batches()[0].to_arrow()

                chalk_logger.info(
                    f"Received {len(arrows)} arrow tables from Snowflake. Total row count is {arrows.num_rows}."
                )
                assert arrows is not None, "Snowflake query"
                return arrows

    def execute_query_to_pyarrow_tables(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
        connection: Optional[Any] = None,
    ) -> Iterator[pa.Table]:
        # this import is safe because the only way we end up here is if we have a valid SnowflakeSource constructed,
        # which already gates this import
        import snowflake.connector
        from sqlalchemy.sql import Select

        if isinstance(finalized_query.query, Select):
            validate_dtypes_for_efficient_execution(finalized_query.query, _supported_sqlalchemy_types_for_pa_querying)

        with (
            snowflake.connector.connect(
                user=self.user,
                account=self.account_identifier,
                password=self.password,
                warehouse=self.warehouse,
                schema=self.schema,
                database=self.db,
            )
            if connection is None
            else contextlib.nullcontext(connection)
        ) as con:
            if con is None:
                chalk_logger.warning("Failed to connect to batch Snowflake")
            assert con is not None
            chalk_logger.info("Established connection with batch Snowflake")
            sql, positional_params, named_params = self.compile_query(finalized_query)
            assert len(positional_params) == 0, "using named param style"
            with con.cursor() as cursor:
                chalk_logger.info("Acquired batch cursor for Snowflake query. Executing.")
                res = cursor.execute(sql, named_params)
                chalk_logger.info("Executed batch Snowflake query. Fetching results.")
                assert res is not None

                chalk_logger.info("Fetching arrow batches from Snowflake.")
                arrows = cursor.fetch_arrow_batches()
                chalk_logger.info("Received arrow batches from Snowflake.")

                if arrows is None:
                    chalk_logger.info("Query returned no results.")
                    return [cursor.get_result_batches()[0].to_arrow()]

                assert arrows is not None
                return arrows

    def execute_query_to_batches(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
        connection: Optional[Any] = None,
    ) -> List[ResultBatch]:
        # this import is safe because the only way we end up here is if we have a valid SnowflakeSource constructed,
        # which already gates this import
        import snowflake.connector
        from sqlalchemy.sql import Select

        if isinstance(finalized_query.query, Select):
            validate_dtypes_for_efficient_execution(finalized_query.query, _supported_sqlalchemy_types_for_pa_querying)

        with (
            snowflake.connector.connect(
                user=self.user,
                account=self.account_identifier,
                password=self.password,
                warehouse=self.warehouse,
                schema=self.schema,
                database=self.db,
            )
            if connection is None
            else contextlib.nullcontext(connection)
        ) as con:
            if con is None:
                chalk_logger.warning("Failed to connect to batch Snowflake")
            assert con is not None
            chalk_logger.info("Established connection with batch Snowflake")
            sql, positional_params, named_params = self.compile_query(finalized_query)
            assert len(positional_params) == 0, "using named param style"
            with con.cursor() as cursor:
                chalk_logger.info("Acquired batch cursor for Snowflake query. Executing.")
                res = cursor.execute(sql, named_params)
                chalk_logger.info("Executed batch Snowflake query. Fetching results.")
                assert res is not None

                chalk_logger.info("Fetching batches from Snowflake.")
                arrows = cursor.get_result_batches()
                chalk_logger.info("Received batches from Snowflake.")
                assert arrows is not None
                return arrows



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, List, Mapping, Optional, Union

import pyarrow as pa

from chalk.features.dataframe import DataFrame
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import polars as pl

from chalk.utils.string import normalize_string_for_matching


def _resolve_implicit_col_feature_mappings_str(
    explicit_mappings: Mapping[str, str],
    expected_features: Iterable[str],
    row_column_names: List[str],
) -> List[Optional[str]]:
    implicit_mappings = {normalize_string_for_matching(f): f for f in expected_features}
    return [
        explicit_mappings.get(col)
        if col in explicit_mappings
        else implicit_mappings.get(normalize_string_for_matching(col), None)
        for col in row_column_names
    ]


def query_arrow_table(query: str, table: pa.Table) -> pa.Table:
    try:
        import duckdb
        import sqlglot
        import sqlglot.expressions
    except ModuleNotFoundError:
        raise missing_dependency_exception("chalkpy[runtime]")
    table_names = list({t.name for t in sqlglot.parse_one(query).find_all(sqlglot.expressions.Table)})
    con = duckdb.connect(database=":memory:")
    if len(table_names) > 1:
        raise ValueError(f"Query should target only one table. Found: {', '.join(table_names)}")
    if len(table_names) == 1:
        con.register(table_names[0], table)
    result_table = con.execute(query).fetch_arrow_table()
    return result_table


def query_as_features(
    query: str,
    fqn_to_name: Mapping[str, str],
    table: pa.Table,
) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    res = query_arrow_table(query, table)
    col_feature_mappings = _resolve_implicit_col_feature_mappings_str(
        # Note: snowflake by default capitalizes all table and column names, although it's
        # technically possible to disable this behavior by quoting things when creating your schema
        # explicit_mappings={k.upper(): v for k, v in self._fields_normalized().items()},
        explicit_mappings={},
        row_column_names=res.column_names,
        expected_features=fqn_to_name.values(),
    )
    name_to_fqn = {v: k for k, v in fqn_to_name.items()}
    df = pl.from_arrow(res)
    assert isinstance(df, pl.DataFrame)
    renamed = df.drop([col for col, f in zip(res.column_names, col_feature_mappings) if f is None]).rename(
        {col: name_to_fqn[f] for col, f in zip(res.column_names, col_feature_mappings) if f is not None}
    )
    return renamed


def query_as_feature_formatted(
    formatted_query: str,
    fqn_to_name: Mapping[str, str],
    table: Union[pa.Table, DataFrame, pl.DataFrame, pl.LazyFrame],
) -> pl.DataFrame:
    # We're going to string replace whatever was there before
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")

    unformatted = formatted_query.replace(str(table), "__chalk_table__")
    pa_table = table
    if isinstance(table, DataFrame):
        pa_table = table.to_polars().collect().to_arrow()
    elif isinstance(table, pl.DataFrame):
        pa_table = table.to_arrow()
    elif isinstance(table, pl.LazyFrame):
        pa_table = table.collect().to_arrow()

    if not isinstance(pa_table, pa.Table):
        raise ValueError(f"Unexpected table type: {type(pa_table)}")

    return query_as_features(unformatted, fqn_to_name, pa_table)
